@@ -1,10 +1,13 @@
 """Python driver for Alicat mass flow controllers, using serial communication.
 
 Distributed under the GNU General Public License v2
+Copyright (C) 2024 Alex Ruddick
 Copyright (C) 2023 NuMat Technologies
 """
+from __future__ import annotations
+
 import asyncio
-from typing import Any, Dict, Optional, Union
+from typing import Any, ClassVar
 
 from .util import Client, SerialClient, TcpClient, _is_float
 
@@ -19,14 +22,13 @@ class FlowMeter:
     connection using pyserial, or an Ethernet <-> serial converter.
     """
 
-    # A dictionary that maps port names to a tuple of connection
-    # objects and the refcounts
-    open_ports: Dict[int, tuple] = {}
-    gases = ['Air', 'Ar', 'CH4', 'CO', 'CO2', 'C2H6', 'H2', 'He',
-             'N2', 'N2O', 'Ne', 'O2', 'C3H8', 'n-C4H10', 'C2H2',
-             'C2H4', 'i-C2H10', 'Kr', 'Xe', 'SF6', 'C-25', 'C-10',
-             'C-8', 'C-2', 'C-75', 'A-75', 'A-25', 'A1025', 'Star29',
-             'P-5']
+    # mapping of port names to a tuple of Client objects and their refcounts
+    open_ports: ClassVar[dict[str, tuple[Client, int]]] = {}
+    gases: ClassVar[list] = ['Air', 'Ar', 'CH4', 'CO', 'CO2', 'C2H6', 'H2', 'He',
+                             'N2', 'N2O', 'Ne', 'O2', 'C3H8', 'n-C4H10', 'C2H2',
+                             'C2H4', 'i-C2H10', 'Kr', 'Xe', 'SF6', 'C-25', 'C-10',
+                             'C-8', 'C-2', 'C-75', 'A-75', 'A-25', 'A1025', 'Star29',
+                             'P-5']
 
     def __init__(self, address: str = '/dev/ttyUSB0', unit: str = 'A', **kwargs: Any) -> None:
         """Connect this driver with the appropriate USB / serial port.
@@ -36,7 +38,14 @@ class FlowMeter:
             unit: The Alicat-specified unit ID, A-Z. Default 'A'.
         """
         if address.startswith('/dev') or address.startswith('COM'):  # serial
-            self.hw: Client = SerialClient(address=address, **kwargs)
+            if address in FlowMeter.open_ports:
+                # Reuse existing connection
+                self.hw, refcount = FlowMeter.open_ports[address]
+                FlowMeter.open_ports[address] = (self.hw, refcount + 1)
+            else:
+                # Open a new connection and store it
+                self.hw: Client = SerialClient(address=address, **kwargs)  # type: ignore[no-redef]
+                FlowMeter.open_ports[address] = (self.hw, 1)
         else:
             self.hw = TcpClient(address=address, **kwargs)
 
@@ -44,14 +53,15 @@ class FlowMeter:
         self.keys = ['pressure', 'temperature', 'volumetric_flow', 'mass_flow',
                      'setpoint', 'gas']
         self.open = True
-        self.firmware: Union[str, None] = None
+        self.firmware: str | None = None
 
-    async def __aenter__(self, *args: Any) -> 'FlowMeter':
+    async def __aenter__(self, *args: Any) -> FlowMeter:
         """Provide async enter to context manager."""
         return self
 
     async def __aexit__(self, *args: Any) -> None:
         """Provide async exit to context manager."""
+        await self.close()
         return
 
     @classmethod
@@ -95,7 +105,7 @@ class FlowMeter:
             raise OSError(f"The FlowMeter with unit ID {self.unit} and "
                            "port {self.hw.address} is not open")
 
-    async def _write_and_read(self, command: str) -> Optional[str]:
+    async def _write_and_read(self, command: str) -> str | None:
         """Wrap the communicator request, to call _test_controller_open() before any request."""
         self._test_controller_open()
         return await self.hw._write_and_read(command)
@@ -124,9 +134,9 @@ class FlowMeter:
         spl = line.split()
         unit, values = spl[0], spl[1:]
 
-        # Mass/volume over range error.
+        # Over range errors for mass, volume, pressure, and temperature
         # Explicitly silenced because I find it redundant.
-        while values[-1].upper() in ['MOV', 'VOV', 'POV']:
+        while values[-1].upper() in ['MOV', 'VOV', 'POV', 'TOV']:
             del values[-1]
         if unit != self.unit:
             raise ValueError("Flow controller unit ID mismatch.")
@@ -142,9 +152,9 @@ class FlowMeter:
         elif len(values) == 2 and len(self.keys) == 6:
             self.keys.insert(1, 'setpoint')
         return {k: (float(v) if _is_float(v) else v)
-                for k, v in zip(self.keys, values)}
-
-    async def set_gas(self, gas: Union[str, int]) -> None:
+                for k, v in zip(self.keys, values, strict=True)}  # type: ignore[call-overload]
+                                                                  # PEP618
+    async def set_gas(self, gas: str | int) -> None:
         """Set the gas type.
 
         Args:
@@ -275,12 +285,19 @@ class FlowMeter:
     async def close(self) -> None:
         """Close the flow meter. Call this on program termination.
 
-        Also closes the serial port if no other FlowMeter object has
+        Also close the serial port if no other FlowMeter object has
         a reference to the port.
         """
         if not self.open:
             return
-        await self.hw.close()
+        port = self.hw.address
+        if port in FlowMeter.open_ports:
+            connection, refcount = FlowMeter.open_ports[port]
+            if refcount > 1:
+                FlowMeter.open_ports[port] = (connection, refcount - 1)
+            else:
+                await connection.close()  # Close the port if no other instance uses it
+                del FlowMeter.open_ports[port]
         self.open = False
 
 
@@ -297,9 +314,9 @@ class FlowController(FlowMeter):
     that the "Input" option is set to "Serial".
     """
 
-    registers = {'mass flow': 0b00100101, 'vol flow': 0b00100100,
-                 'abs pressure': 0b00100010, 'gauge pressure': 0b00100110,
-                 'diff pressure': 0b00100111}
+    registers: ClassVar[dict] = {'mass flow': 0b00100101, 'vol flow': 0b00100100,
+                                 'abs pressure': 0b00100010, 'gauge pressure': 0b00100110,
+                                 'diff pressure': 0b00100111}
 
     def __init__(self, address: str='/dev/ttyUSB0', unit: str='A', **kwargs: Any) -> None:
         """Connect this driver with the appropriate USB / serial port.
@@ -314,11 +331,11 @@ class FlowController(FlowMeter):
             self.control_point = await self._get_control_point()
         self._init_task = asyncio.create_task(_init_control_point())
 
-    async def __aenter__(self, *args: Any) -> 'FlowController':
+    async def __aenter__(self, *args: Any) -> FlowController:
         """Provide async enter to context manager."""
         return self
 
-    async def _write_and_read(self, command: str) -> Optional[str]:
+    async def _write_and_read(self, command: str) -> str | None:
         """Wrap the communicator request.
 
         (1) Ensure _init_task is called once before the first request
@@ -346,8 +363,6 @@ class FlowController(FlowMeter):
             The state of the flow controller, as a dictionary.
         """
         state = await super().get()
-        if state is None:
-            return None
         state['control_point'] = self.control_point
         return state
 
@@ -374,8 +389,52 @@ class FlowController(FlowMeter):
             await self._set_control_point('abs pressure')
         await self._set_setpoint(pressure)
 
+    async def get_totalizer_batch(self, batch: int = 1) -> str:
+        """Get the totalizer batch volume (firmware 10v00).
+
+        Args:
+            batch: Which of the two totalizer batches to query.
+                Default is 1; some devices have 2
+
+        Returns:
+            line: Current value of totalizer batch
+        """
+        command = f'{self.unit}$$TB {batch}'
+        line = await self._write_and_read(command)
+
+        if line == '?':
+            raise OSError("Unable to read totalizer batch volume.")
+        values = line.split(" ")  # type: ignore[union-attr]
+        return f'{values[2]} {values[4]}' # returns 'batch vol' 'units'
+
+    async def set_totalizer_batch(self, batch_volume: float, batch: int = 1, units: str = 'default') -> None:
+        """Set the totalizer batch volume (firmware 10v00).
+
+        Args:
+            batch: Which of the two totalizer batches to set.
+                Default is 1; some devices have 2
+            batch_volume: Target batch volume, in same units as units
+                on device
+            units: Units of the volume being provided. Default
+                is 0, so device returns default engineering units.
+        """
+        engineering_units_table = {"default":0, "SμL":2, "SmL":3, "SL":4, \
+                    "Scm3":6, "Sm3":7, "Sin3":8, "Sft3":9, "kSft3":10, "NμL":32, \
+                    "NmL":33, "NL":34, "Ncm3":36, "Nm3":37}
+
+        if units in engineering_units_table:
+            units_no = engineering_units_table[units]
+        else:
+            raise ValueError("Units not in unit list. Please consult Appendix B-3 of the Alicat Serial Primer.")
+
+        command = f'{self.unit}$$TB {batch} {batch_volume} {units_no}'
+        line = await self._write_and_read(command)
+
+        if line == '?':
+            raise OSError("Unable to set totalizer batch volume. Check if volume is out of range for device.")
+
     async def hold(self) -> None:
-        """Override command to issue a valve hold.
+        """Override command to issue a valve hold (firmware 5v07).
 
         For a single valve controller, hold the valve at the present value.
         For a dual valve flow controller, hold the valve at the present value.
@@ -414,12 +473,13 @@ class FlowController(FlowMeter):
             pid_values.append(value_spl[3])
 
         return {k: (v if k == self.pid_keys[-1] else str(v))
-                for k, v in zip(self.pid_keys, pid_values)}
+                for k, v in zip(self.pid_keys, pid_values, strict=True)}  # type: ignore[call-overload]
+                                                                          # PEP618
 
-    async def set_pid(self, p: Optional[int]=None,
-                            i: Optional[int]=None,
-                            d: Optional[int]=None,
-                            loop_type: Optional[str]=None) -> None:
+    async def set_pid(self, p: int | None=None,
+                            i: int | None=None,
+                            d: int | None=None,
+                            loop_type: str | None=None) -> None:
         """Set specified PID parameters.
 
         Args:
@@ -462,7 +522,17 @@ class FlowController(FlowMeter):
         except IndexError:
             current = None
         if current is not None and abs(current - setpoint) > 0.01:
-            raise OSError("Could not set setpoint.")
+            # possibly the setpoint is being ramped
+            command = f'{self.unit}LS'
+            line = await self._write_and_read(command)
+            if not line:
+                raise OSError("Could not set setpoint.")
+            try:
+                commanded = float(line.split()[2])
+            except IndexError:
+                raise OSError("Could not set setpoint.") from None
+            if commanded is not None and abs(commanded - setpoint) > 0.01:
+                raise OSError("Could not set setpoint.")
 
     async def _get_control_point(self) -> str:
         """Get the control point, and save to internal variable."""
@@ -495,3 +565,45 @@ class FlowController(FlowMeter):
         if value != reg:
             raise OSError("Could not set control point.")
         self.control_point = point
+
+    async def set_ramp_config(self, config: dict[str, bool]) -> None:
+        """Configure the setpoint ramp settings (firmware 10v05).
+
+        `up`: whether the controller ramps when increasing the setpoint,
+        `down`: whether the controller ramps when decreasing the setpoint,
+                (this includes setpoints below 0 on bidirectional devices),
+        `zero`: whether the controller ramps when establishing a zero setpoint,
+        `power`: whether the controller ramps when using a power-up setpoint
+        """
+        command = (f"{self.unit}LSRC"
+                  f" {1 if config['up'] else 0}"
+                  f" {1 if config['down'] else 0}"
+                  f" {1 if config['zero'] else 0}"
+                  f" {1 if config['power'] else 0}")
+        line = await self._write_and_read(command)
+        if not line or self.unit not in line:
+            raise OSError("Could not set ramp config.")
+
+
+    async def get_ramp_config(self) -> dict[str, bool]:
+        """Get the setpoint ramp settings (firmware 10v05).
+
+        `up`: whether the controller ramps when increasing the setpoint,
+        `down`: whether the controller ramps when decreasing the setpoint,
+                (this includes setpoints below 0 on bidirectional devices),
+        `zero`: whether the controller ramps when establishing a zero setpoint,
+        `power`: whether the controller ramps when using a power-up setpoint
+        """
+        command = f"{self.unit}LSRC"
+        line = await self._write_and_read(command)
+        if not line or self.unit not in line:
+            raise OSError("Could not read ramp config.")
+        values = line[2:].split(' ')
+        if len(values) != 4:
+            raise OSError("Could not read ramp config.")
+        return {
+            'up': values[0] == '1',
+            'down': values[1] == '1',
+            'zero': values[2] == '1',
+            'power': values[3] == '1',
+        }
